﻿namespace MigraDocDemo.BusinessLogic
{
    using MigraDoc.DocumentObjectModel;
    using MigraDoc.DocumentObjectModel.Tables;
    using MigraDoc.Rendering;
    using PdfSharp.Pdf;
    using System.Collections.Generic;
    using System.IO;

    using Models;
    using Enums;

    public class PdfService
    {
        private readonly Color HeartsAlikeMainColor = new Color(255, 0, 110);
        private readonly Color HeartsAlikeSupportColor = new Color(255, 0, 110);

        /// <summary>
        /// Creates the PDF document and saves it.
        /// </summary>
        /// <param name="model">The Hearts Alike model to generate a PDF document for.</param>
        /// <returns>The full filename of the generated PDF document.</returns>
        public string CreatePdf(HeartsAlikeModel model)
        {
            var document = CreateDocument();

            RenderModel(document.Sections[0], model);

            var filePath = SaveDocument(document, "HeartsAlike.pdf");

            return filePath;
        }

        /// <summary>
        /// Creates a new PDF document, sets styles and renders document-wide content like header, footer and logo.)
        /// </summary>
        /// <param name="title">The reportNumber of the Corff this PDF document wil contain.</param>
        /// <param name="author">The login name of the author of the document.</param>
        /// <returns></returns>
        private Document CreateDocument()
        {
            var document = new Document();

            document.Info.Title = "Hearts Alike";
            document.Info.Subject = "Shows some advanced features of MigraDoc";
            document.Info.Author = "Hearts Alike Inc.";

            DefineStyles(document);

            var section = document.AddSection();

            DefinePageSetup(section);
            RenderLogo(section);
            RenderTitle(section);
            RenderSideBar(section);
            RenderFooter(section);

            return document;
        }

        /// <summary>
        /// Defines new styles to use in the document as well as changes to the default styles.
        /// </summary>
        /// <param name="section">The PDF document to change the styles for.</param>
        private void DefineStyles(Document document)
        {
            // Get the predefined style Normal.
            Style style = document.Styles[StyleNames.Normal];
            // Because all styles are derived from Normal, the next line changes the 
            // font of the whole document. Or, more exactly, it changes the font of
            // all styles and paragraphs that do not redefine the font.
            style.Font.Name = "Comic Sans MS";
            style.Font.Size = 16;

            style = document.AddStyle("Bold", "Normal");
            style.Font.Name = "Comic Sans MS";
            style.Font.Bold = true;

            style = document.AddStyle("BulletList", "Normal");
            style.ParagraphFormat.FirstLineIndent = Unit.FromCentimeter(-0.5);
            style.ParagraphFormat.LeftIndent = Unit.FromCentimeter(0.5);
            style.ParagraphFormat.ListInfo = new ListInfo { ListType = ListType.BulletList1 };

            style = document.Styles[StyleNames.Heading1];
            style.Font.Name = "Snap ITC Regular";
            style.Font.Color = HeartsAlikeMainColor;
            style.Font.Size = 40;

            style = document.Styles[StyleNames.Header];
            style.ParagraphFormat.AddTabStop(Unit.FromCentimeter(15.7), TabAlignment.Right);
        }

        /// <summary>
        /// Defines the page settings like margins.
        /// </summary>
        /// <param name="section">The section of the PDF to define the page settings for.</param>
        private void DefinePageSetup(Section section)
        {
            section.PageSetup.PageFormat = PageFormat.A4;
            section.PageSetup.PageHeight = Unit.FromCentimeter(29.7);
            section.PageSetup.PageWidth = Unit.FromCentimeter(21);
            section.PageSetup.Orientation = Orientation.Portrait;
            section.PageSetup.LeftMargin = Unit.FromCentimeter(3.3);
            section.PageSetup.RightMargin = Unit.FromCentimeter(2.0);
            section.PageSetup.TopMargin = Unit.FromCentimeter(6.5);
            section.PageSetup.BottomMargin = Unit.FromCentimeter(2.0);
            section.PageSetup.FooterDistance = Unit.FromCentimeter(1.25);
            section.PageSetup.HeaderDistance = Unit.FromCentimeter(1.25);
            section.PageSetup.OddAndEvenPagesHeaderFooter = false;
            section.PageSetup.DifferentFirstPageHeaderFooter = false;
            section.PageSetup.SectionStart = BreakType.BreakNextPage;
            section.PageSetup.MirrorMargins = false;
            section.PageSetup.HorizontalPageBreak = false;
        }

        /// <summary>
        /// Render a logo in primary header.
        /// </summary>
        /// <param name="section">The section of the PDF to render into.</param>
        private void RenderLogo(Section section)
        {
            var rootPath = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
            var logoFilename = Path.Combine(rootPath, "Images", "HeartsAlike.png");

            var image = section.Headers.Primary.AddImage(logoFilename);
            image.RelativeHorizontal = MigraDoc.DocumentObjectModel.Shapes.RelativeHorizontal.Margin;
            image.RelativeVertical = MigraDoc.DocumentObjectModel.Shapes.RelativeVertical.Page;
            image.Left = Unit.FromCentimeter(0);
            image.Top = Unit.FromCentimeter(1);
            image.LockAspectRatio = true;
        }

        /// <summary>
        /// Render a title in primary header.
        /// </summary>
        /// <param name="section">The section of the PDF to render into.</param>
        private void RenderTitle(Section section)
        {
            var textFrame = section.Headers.Primary.AddTextFrame();
            textFrame.RelativeHorizontal = MigraDoc.DocumentObjectModel.Shapes.RelativeHorizontal.Margin;
            textFrame.RelativeVertical = MigraDoc.DocumentObjectModel.Shapes.RelativeVertical.Page;
            textFrame.Left = Unit.FromCentimeter(0);
            textFrame.Top = Unit.FromCentimeter(2);
            textFrame.Width = Unit.FromCentimeter(15.7);
            var title = textFrame.AddParagraph("Hearts Alike");
            title.Style = "Heading1";
            title.Format.Alignment = ParagraphAlignment.Right;
        }

        /// <summary>
        /// Render vertical colored bar on left side of page.
        /// </summary>
        /// <param name="section">The section of the PDF to render into.</param>
        private void RenderSideBar(Section section)
        {
            var textFrame = section.AddTextFrame();

            textFrame.RelativeHorizontal = MigraDoc.DocumentObjectModel.Shapes.RelativeHorizontal.Page;
            textFrame.RelativeVertical = MigraDoc.DocumentObjectModel.Shapes.RelativeVertical.Page;
            textFrame.Top = 0;
            textFrame.Left = 0;
            textFrame.Height = Unit.FromCentimeter(29.7);
            textFrame.Width = Unit.FromCentimeter(2.1);
            textFrame.FillFormat.Color = HeartsAlikeMainColor;

            var par = textFrame.AddParagraph();
            par.Format.ClearAll();
            par.Comment = "Sidebar";
        }

        /// <summary>
        /// Renders primary footer.
        /// </summary>
        /// <param name="section">The section of the PDF to render into.</param>
        private void RenderFooter(Section section)
        {
            Paragraph paragraph = section.Footers.Primary.AddParagraph();
            paragraph.AddText("Copyright © 2016, HeartsAlike");
            paragraph.Format.Font.Size = 9;
            paragraph.Format.Alignment = ParagraphAlignment.Center;
        }

        /// <summary>
        /// Render the Hearts Alike model to the PDF section.
        /// </summary>
        /// <param name="section">The section of the PDF to render into.</param>
        /// <param name="model">The model to render</param>
        private void RenderModel(Section section, HeartsAlikeModel model)
        {
            var usableWidth = (int)(section.PageSetup.PageWidth - section.PageSetup.LeftMargin - section.PageSetup.RightMargin);
            var leftPanelWidth = (int)(usableWidth * 0.50);
            var rightPanelWidth = usableWidth - leftPanelWidth;

            // construct panels using table
            var table = section.AddTable();
            table.TopPadding = Unit.FromMillimeter(1.2);
            table.BottomPadding = Unit.FromMillimeter(1.2);
            table.LeftPadding = Unit.FromMillimeter(1.2);
            table.RightPadding = Unit.FromMillimeter(1.2);

            table.Borders.Style = BorderStyle.None;
            table.Borders.Width = 0;

            table.Rows.LeftIndent = 0;

            // Define panels
            Column column = table.AddColumn();
            column.Width = leftPanelWidth;
            column.RightPadding = Unit.FromMillimeter(3);
            column.Format.Alignment = ParagraphAlignment.Left;

            column.Borders.Top.Color = HeartsAlikeMainColor;
            column.Borders.Top.Style = BorderStyle.Single;
            column.Borders.Top.Width = 2;
            column.Borders.Right.Color = HeartsAlikeMainColor;
            column.Borders.Right.Style = BorderStyle.Single;
            column.Borders.Right.Width = 2;

            column = table.AddColumn();
            column.Width = rightPanelWidth;
            column.RightPadding = Unit.FromMillimeter(2);
            column.Format.Alignment = ParagraphAlignment.Left;

            column.Borders.Top.Color = HeartsAlikeMainColor;
            column.Borders.Top.Style = BorderStyle.Single;
            column.Borders.Top.Width = 2;
            column.Borders.Left.Color = HeartsAlikeMainColor;
            column.Borders.Left.Style = BorderStyle.Single;
            column.Borders.Left.Width = 2;

            // Create the two panels (i.e. a row with two columns)
            Row row = table.AddRow();

            // Create table for actual content within the left panel
            //Cell cell = row.Cells[0];
            //Table innerTableLeft = cell.AddTable(); // <- Cell does not have a method AddTable()

            RenderLeftPanel(row.Cells[0], model);
            RenderRightPanel(row.Cells[1], model);
        }

        /// <summary>
        /// Renders the left panel ('I am').
        /// </summary>
        /// <param name="cell">The cell (panel) to render the contents</param>
        /// <param name="model">the model containing all Hearts Alike data.</param>
        private void RenderLeftPanel(Cell cell, HeartsAlikeModel model)
        {
            cell.Comment = "LeftPanel";

            var usableWidth = (int)(cell.Column.Width - cell.Column.LeftPadding - cell.Column.RightPadding - cell.Column.Borders.Width - cell.Column.Borders.Width - cell.Column.Borders.DistanceFromLeft - cell.Column.Borders.DistanceFromRight);
            var leftColumnWidth = (int)(usableWidth * 0.3);
            var rightColumnWidth = usableWidth - leftColumnWidth;

            // Create table for actual content within the left panel
            //Table innerTableLeft = cell.AddTable(); // <- Cell does not have a method AddTable()

            // Nifty trick to get nested table
            var table = cell.Elements.AddTable();

            table.LeftPadding = 0;
            table.RightPadding = 0;
            table.Rows.LeftIndent = 0;

            // Uncomment if you'd like to see the inner table
            //table.Borders.Color = HeartsAlikeSupportColor;
            //table.Borders.Style = BorderStyle.Dot;
            //table.Borders.Width = 1;

            Column column = table.AddColumn();
            column.Width = leftColumnWidth;
            column.LeftPadding = Unit.FromMillimeter(2.5);
            column.RightPadding = Unit.FromMillimeter(2.5);
            column.Format.Alignment = ParagraphAlignment.Left;

            column = table.AddColumn();
            column.Width = rightColumnWidth;
            column.LeftPadding = Unit.FromMillimeter(2.5);
            column.RightPadding = Unit.FromMillimeter(2.5);
            column.Format.Alignment = ParagraphAlignment.Right;

            Row row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("I am:").Style = "Bold";
            row.Cells[0].MergeRight = 1;

            row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("Gender").Style = "Bold";
            row.Cells[1].AddParagraph(model.MyGender.ToString());

            row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("Age").Style = "Bold";
            row.Cells[1].AddParagraph(model.MyAge.ToString());

            row = table.AddRow();
            row.BottomPadding = 5;
            RenderAppearanceTraits(row.Cells[0], model.MyAppearanceTraits);
            row.Cells[0].MergeRight = 1;

            row = table.AddRow();
            row.BottomPadding = 5;
            RenderCharacterTraits(row.Cells[0], model.MyCharacterTraits);
            row.Cells[0].MergeRight = 1;
        }

        /// <summary>
        /// Renders the left panel ('I am').
        /// </summary>
        /// <param name="cell">The cell (panel) to render the contents</param>
        /// <param name="model">the model containing all Hearts Alike data.</param>
        private void RenderRightPanel(Cell cell, HeartsAlikeModel model)
        {
            cell.Comment = "RightPanel";

            var usableWidth = (int)(cell.Column.Width - cell.Column.LeftPadding - cell.Column.RightPadding - cell.Column.Borders.Width - cell.Column.Borders.Width - cell.Column.Borders.DistanceFromLeft - cell.Column.Borders.DistanceFromRight);
            var leftColumnWidth = (int)(usableWidth * 0.3);
            var rightColumnWidth = usableWidth - leftColumnWidth;

            // Nifty trick to get nested table
            var table = cell.Elements.AddTable();

            table.LeftPadding = 0;
            table.RightPadding = 0;
            table.Rows.LeftIndent = 0;

            // Uncomment if you'd like to see the inner table
            //table.Borders.Color = HeartsAlikeSupportColor;
            //table.Borders.Style = BorderStyle.Dot;
            //table.Borders.Width = 1;

            Column column = table.AddColumn();
            column.Width = leftColumnWidth;
            column.LeftPadding = Unit.FromMillimeter(2.5);
            column.RightPadding = Unit.FromMillimeter(2.5);
            column.Format.Alignment = ParagraphAlignment.Left;

            column = table.AddColumn();
            column.Width = rightColumnWidth;
            column.LeftPadding = Unit.FromMillimeter(2.5);
            column.RightPadding = Unit.FromMillimeter(2.5);
            column.Format.Alignment = ParagraphAlignment.Right;

            Row row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("I am looking for:").Style = "Bold";
            row.Cells[0].MergeRight = 1;

            row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("Gender").Style = "Bold";
            row.Cells[1].AddParagraph(model.LookingForGender.ToString());

            row = table.AddRow();
            row.BottomPadding = 5;
            row.Cells[0].AddParagraph("Age").Style = "Bold";
            row.Cells[1].AddParagraph(string.Format("{0} ± {1}", model.LookingForAge.ToString(), model.MaxAgeDifference));

            row = table.AddRow();
            row.BottomPadding = 5;
            RenderAppearanceTraits(row.Cells[0], model.LookingForAppearanceTraits);
            row.Cells[0].MergeRight = 1;

            row = table.AddRow();
            row.BottomPadding = 5;
            RenderCharacterTraits(row.Cells[0], model.LookingForCharacterTraits);
            row.Cells[0].MergeRight = 1;
        }

        /// <summary>
        /// Render all appearance traits as a bullet list.
        /// </summary>
        /// <param name="cell">The cell to render to.</param>
        /// <param name="traits">The list of traits to render.</param>
        private void RenderAppearanceTraits(Cell cell, IEnumerable<AppearanceTraits> traits)
        {
            cell.AddParagraph("Appearance").Style = "Bold";

            foreach (var trait in traits)
            {
                cell.AddParagraph(trait.ToString()).Style = "BulletList";
            }
        }

        /// <summary>
        /// Render all character traits as a bullet list.
        /// </summary>
        /// <param name="cell">The cell to render to.</param>
        /// <param name="traits">The list of traits to render.</param>
        private void RenderCharacterTraits(Cell cell, IEnumerable<CharacterTraits> traits)
        {
            cell.AddParagraph("Character").Style = "Bold";

            foreach (var trait in traits)
            {
                cell.AddParagraph(trait.ToString()).Style = "BulletList";
            }
        }

        /// <summary>
        /// Saves the document to disk.
        /// </summary>
        /// <param name="document">The PDF Document to save.</param>
        /// <param name="filename">The filename to use.</param>
        /// <returns>The full filename of the generated PDF document.</returns>
        private string SaveDocument(Document document, string filename)
        {
            var renderer = new PdfDocumentRenderer(true, PdfFontEmbedding.Always);
            var rootPath = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
            var filePath = Path.Combine(rootPath, "Output");

            if (!Directory.Exists(filePath))
            {
                Directory.CreateDirectory(filePath);
            }

            var fullName = Path.Combine(filePath, filename);

            renderer.Document = document;

#if DEBUG
            // Write the MigraDoc document definition language file to disk for debug purposes
            var mdddlFilePath = fullName + ".mdddl";
            MigraDoc.DocumentObjectModel.IO.DdlWriter dw = new MigraDoc.DocumentObjectModel.IO.DdlWriter(mdddlFilePath);
            dw.WriteDocument(document);
            dw.Close();
#endif

            renderer.RenderDocument();
            renderer.PdfDocument.Save(fullName);

            return fullName;
        }
    }
}
